# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import numpy as np

from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces import ConfigReaderBehaviorModels
from bark.runtime.scenario.interaction_dataset_processing import BehaviorFromTrack

from bark.core.models.behavior import *
from bark.runtime.commons.parameters import ParameterServer

  # this config reader defines behavior models with fixed type for all agents
class FixedBehaviorType(ConfigReaderBehaviorModels):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.param_servers = []

  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    model_type = config_param_object["ModelType", "Type of behavior model \
                used for all vehicles", "BehaviorIDMClassic"]
    model_params = config_param_object.AddChild("ModelParams")
    self.param_servers.append(model_params) # use the same param server for all models
    behavior_models = []
    behavior_model_types = []
    for _ in agent_states:
      bark_model, _ = self.model_from_model_type(model_type, model_params)
      behavior_models.append(bark_model)
      behavior_model_types.append(model_type)
    return behavior_models, {"behavior_model_types" : behavior_model_types}, config_param_object

  def model_from_model_type(self, model_type, params):
    bark_model = eval("{}(params)".format(model_type))
    return bark_model, params

  def get_param_servers(self):
    return self.param_servers


class InteractionDataBehaviors(ConfigReaderBehaviorModels):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.param_servers = []

  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    tracks = kwargs["tracks"]
    xy_offset = kwargs["xy_offset"]
    start_time = kwargs["start_time"]
    end_time = kwargs["end_time"]

    behavior_models = []
    behavior_model_types = []

    for idx, _ in enumerate(agent_states):
      track = tracks[idx]
      params = ParameterServer()
      behavior = BehaviorFromTrack(track, params, xy_offset, start_time, end_time)
      self.param_servers.append(params)
      behavior_models.append(behavior)
      behavior_model_types.append("BehaviorStaticTrajectory")
    return behavior_models, {"behavior_model_types" : behavior_model_types}, config_param_object

  def get_param_servers(self):
    return self.param_servers

class SampleBehaviorType(ConfigReaderBehaviorModels):
  def __init__(self, *args, **kwargs):
    super().__init__( *args, **kwargs)
    self.param_servers = []

  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    model_types = config_param_object["ModelTypesList", "Type of behavior model" \
                "used for all vehicles", ["BehaviorIDMClassic", "BehaviorMobilRuleBased"]]
    model_params = config_param_object.AddChild("ModelParams")
    # ----- DEFAULT PARAMETER HANDLING
    # based on types retrieve default params which are maintained as scenario defaults
    for model_type in model_types:
        behavior_params = model_params.AddChild(model_type)
        _, _ = self.model_from_model_type(model_type, behavior_params)
        #param server must be persisted for each behavior to enable serialization of parameters
        

    #------ BEHAVIOR MODEL SAMPLING
    behavior_models = []
    behavior_model_types = []
    for _ in agent_states:
        model_idx = self.random_state.randint(low=0, high=len(model_types), size=None) 
        model_type = model_types[model_idx]
        model_type_params = model_params.AddChild(model_type)
        params = ParameterServer()
        bark_model, params = self.model_from_model_type(model_type, model_type_params)
        self.param_servers.append(model_type_params) 
        behavior_models.append(bark_model)
        behavior_model_types.append(model_type)
    return behavior_models, {"behavior_model_types" : behavior_model_types}, config_param_object

  def model_from_model_type(self, model_type, params):
    bark_model = eval("{}(params)".format(model_type))    
    return bark_model, params

  def get_param_servers(self):
    return self.param_servers


from .interaction_dataset_reader import BarkStateFromMotionState
from .interaction_dataset_reader import TrajectoryFromTrack
from .interaction_dataset_reader import ShapeFromTrack
from .interaction_dataset_reader import InitStateFromTrack
from .interaction_dataset_reader import GoalDefinitionFromTrack
from .interaction_dataset_reader import BehaviorFromTrack
from .interaction_dataset_reader import InteractionDatasetReader
from .dataset_decomposer import  DatasetDecomposer

__all__ = ["DatasetDecomposer",
          "InteractionDatasetReader",
          "BarkStateFromMotionState",
          "TrajectoryFromTrack",
          "ShapeFromTrack",
          "InitStateFromTrack",
          "GoalDefinitionFromTrack",
          "BehaviorFromTrack"]



from setuptools import setup, find_packages, Extension
import sys, os

with open("Readme.md", "r") as fh:
    long_description = fh.read()

# A dummy native extension to mark module as platform specific
ext_modules= []
try:
    os.mkdir('build')
except FileExistsError:
    # directory already exists - is already created by earlier run
    pass
open('build/temp.c','w').close()
temp_ext = Extension('_temp', sources=['build/temp.c'])
ext_modules.append(temp_ext)

setup(
    name = "bark-ml",
    version = "0.2.6",
    description = "Machine Learning Applied to Autonomous Driving",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers = ["Development Status :: 4 - Beta",
                   "Intended Audience :: Science/Research",
                   "License :: OSI Approved :: MIT License",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python :: 3.7"],
    keywords = "simulator autonomous driving machine learning",
    url = "https://github.com/bark-simulator/bark-ml",
    author = "Patrick Hart, Julian Bernhard, Klemens Esterle, Tobias Kessler",
    author_email = "patrickhart.1990@gmail.com",
    license = "MIT",
    packages=find_packages(),
    install_requires=[
        'pygame>=1.9.6',
        'gym>=0.17.2',
        'tensorflow>=2.2.0',
        'tensorboard>=2.2.2',
        'tf-agents>=0.5.0',
        'tensorflow-probability>=0.10.0',
        'bark-simulator>=1.0.0'
    ],
    ext_modules=ext_modules,
    test_suite='nose.collector',
    tests_require=['nose'],
    include_package_data=True,
    zip_safe=False,
    python_requires='>=3.7',
)



# Copyright (c) 2020 fortiss GmbH
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

from bark.runtime.scenario.interaction_dataset_processing.agent_track_info import AgentTrackInfo


class ScenarioTrackInfo:
    def __init__(self, track_filename, ego_track_info, xy_offset, start_ts=None, end_ts=None, prec=-2):
        self._track_filename = track_filename
        self._ego_track_info = ego_track_info
        self._xy_offset = xy_offset

        if start_ts is None:
            self._start_ts = int(round(ego_track_info.GetStartTimeMs(), prec))
        else:
            self._start_ts = int(round(start_ts, prec))

        if end_ts is None:
            self._end_ts = int(round(ego_track_info.GetEndTimeMs(), prec))
        else:
            self._end_ts = int(round(end_ts, prec))

        self._other_agents_track_infos = {}

    def __str__(self):
        str_out = 'start_ts={}ms end_ts={}ms ego: {}'.format(
            self.GetStartTimeMs(), self.GetEndTimeMs(), self._ego_track_info)
        for other in self.GetOtherTrackInfos().values():
            str_out = str_out + ', other: {}'.format(other)
        return str_out

    def AddTrackInfoOtherAgent(self, track_info_other):
        track_id = track_info_other.GetTrackId()
        self._other_agents_track_infos[track_id] = track_info_other
        if track_info_other.GetStartTimeMs() > self.GetEndTimeMs():
            raise ValueError("Agent cannot start after scenario end")

    def GetTrackFilename(self):
        return self._track_filename

    def GetEgoTrackInfo(self):
        return self._ego_track_info

    def GetXYOffset(self):
        return self._xy_offset

    def GetStartTimeMs(self):
        # time in ms
        return self._start_ts

    def GetEndTimeMs(self):
        # time in ms
        return self._end_ts

    def GetOtherTrackInfos(self):
        return self._other_agents_track_infos

    def GetTimeOffsetOfAgentInSec(self, agent_id):
        # Time Offset when Agent becomes valid in relation to start time of scenario
        if (agent_id == self._ego_track_info.GetTrackId()):
            return 0.0
        else:
            start_ts = self._other_agents_track_infos[agent_id].GetStartTimeMs(
            )
            timestamp_offset = float(start_ts - self.GetStartTimeMs()) / 1000.0
            return timestamp_offset

    def TimeSanityCheck(self):
        if self.GetStartTimeMs() < self.GetEgoTrackInfo().GetStartTimeMs():
            raise ValueError("ego agent {} starts after scenario".format(
                self.GetEgoTrackInfo().GetTrackId()))
        elif self.GetEndTimeMs() > self.GetEgoTrackInfo().GetEndTimeMs():
            raise ValueError("ego agent {} ends before scenario".format(
                self.GetEgoTrackInfo().GetTrackId()))
        else:
            return True



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import tensorflow as tf
import numpy as np

# BARK
from bark.core.models.behavior import BehaviorModel, BehaviorDynamicModel

# tfa
from tf_agents.networks import actor_distribution_network
from tf_agents.networks import normal_projection_network
from tf_agents.agents.ddpg import critic_network
from tf_agents.policies import greedy_policy

from tf_agents.agents.sac import sac_agent
from tf_agents.replay_buffers import tf_uniform_replay_buffer
from tf_agents.utils.common import Checkpointer
from tf_agents.trajectories import time_step as ts

from bark_ml.library_wrappers.lib_tf_agents.agents.tfa_agent import BehaviorTFAAgent
from bark_ml.behaviors.cont_behavior import BehaviorContinuousML


class BehaviorSACAgent(BehaviorTFAAgent):
  """SAC-Agent
     This agent is based on the tf-agents library.
  """
  def __init__(self,
               environment=None,
               params=None,
               observer=None):
    self._sac_params = params["ML"]["BehaviorSACAgent"]
    BehaviorTFAAgent.__init__(
      self,
      environment=environment,
      params=params,
      observer=observer)
    BehaviorModel.__init__(self, params)
    self._replay_buffer = self.GetReplayBuffer()
    self._dataset = self.GetDataset()
    self._collect_policy = self.GetCollectionPolicy()
    self._eval_policy = self.GetEvalPolicy()

  def GetAgent(self, env, params):
    def _normal_projection_net(action_spec, init_means_output_factor=0.1):
      return normal_projection_network.NormalProjectionNetwork(
        action_spec,
        mean_transform=None,
        state_dependent_std=True,
        init_means_output_factor=init_means_output_factor,
        std_transform=sac_agent.std_clip_transform,
        scale_distribution=True)

    # actor network
    actor_net = actor_distribution_network.ActorDistributionNetwork(
        env.observation_spec(),
        env.action_spec(),
        fc_layer_params=tuple(
          self._sac_params["ActorFcLayerParams", "", [512, 256, 256]]),
        continuous_projection_net=_normal_projection_net)

    # critic network
    critic_net = critic_network.CriticNetwork(
      (env.observation_spec(), env.action_spec()),
      observation_fc_layer_params=None,
      action_fc_layer_params=None,
      joint_fc_layer_params=tuple(
        self._sac_params["CriticJointFcLayerParams", "", [512, 256, 256]]))
    
    # agent
    tf_agent = sac_agent.SacAgent(
      env.time_step_spec(),
      env.action_spec(),
      actor_network=actor_net,
      critic_network=critic_net,
      actor_optimizer=tf.compat.v1.train.AdamOptimizer(
          learning_rate=self._sac_params["ActorLearningRate", "", 3e-4]),
      critic_optimizer=tf.compat.v1.train.AdamOptimizer(
          learning_rate=self._sac_params["CriticLearningRate", "", 3e-4]),
      alpha_optimizer=tf.compat.v1.train.AdamOptimizer(
          learning_rate=self._sac_params["AlphaLearningRate", "", 3e-4]),
      target_update_tau=self._sac_params["TargetUpdateTau", "", 0.05],
      target_update_period=self._sac_params["TargetUpdatePeriod", "", 3],
      td_errors_loss_fn=tf.math.squared_difference,
      gamma=self._sac_params["Gamma", "", 0.995],
      reward_scale_factor=self._sac_params["RewardScaleFactor", "", 1.],
      train_step_counter=self._ckpt.step,
      name=self._sac_params["AgentName", "", "sac_agent"],
      debug_summaries=self._sac_params["DebugSummaries", "", False])
    
    tf_agent.initialize()
    return tf_agent

  def GetReplayBuffer(self):
    return tf_uniform_replay_buffer.TFUniformReplayBuffer(
      data_spec=self._agent.collect_data_spec,
      batch_size=self._wrapped_env.batch_size,
      max_length=self._sac_params["ReplayBufferCapacity", "", 100000])

  def GetDataset(self):
    dataset = self._replay_buffer.as_dataset(
      num_parallel_calls=self._sac_params["ParallelBufferCalls", "", 1],
      sample_batch_size=self._sac_params["BatchSize", "", 512],
      num_steps=self._sac_params["BufferNumSteps", "", 2]) \
        .prefetch(self._sac_params["BufferPrefetch", "", 3])
    return dataset

  def GetCollectionPolicy(self):
    return self._agent.collect_policy

  def GetEvalPolicy(self):
    return greedy_policy.GreedyPolicy(self._agent.policy)

  def Reset(self):
    pass
  
  @property
  def collect_policy(self):
    return self._collect_policy

  @property
  def eval_policy(self):
    return self._eval_policy




# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import time
import numpy as np
import logging
import copy
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm

# bark
from bark.runtime.commons.parameters import ParameterServer
from bark.core.models.behavior import BehaviorIDMLaneTracking
from bark.core.world.evaluation import CaptureAgentStates
from bark.runtime.viewer.matplotlib_viewer import MPViewer

# bark-ml
from bark_ml.environments.single_agent_runtime import SingleAgentRuntime
from bark_ml.commons.tracer import Tracer

from matplotlib.colors import LinearSegmentedColormap


class CounterfactualRuntime(SingleAgentRuntime):
  def __init__(self,
               blueprint=None,
               ml_behavior=None,
               observer=None,
               evaluator=None,
               step_time=None,
               viewer=None,
               scenario_generator=None,
               render=False,
               max_col_rate=0.1,
               behavior_model_pool=None,
               ego_rule_based=None,
               params=None):
    SingleAgentRuntime.__init__(
      self,
      blueprint=blueprint,
      ml_behavior=ml_behavior,
      observer=observer,
      evaluator=evaluator,
      step_time=step_time,
      viewer=viewer,
      scenario_generator=scenario_generator,
      render=render)
    self._params = params or ParameterServer()
    self._max_col_rate = params["ML"][
      "MaxColRate",
      "Max. collision rate allowed over all counterfactual worlds.", 0.1]
    self._cf_simulation_steps = params["ML"][
      "CfSimSteps",
      "Simulation steps for the counterfactual worlds.", 5]
    self._visualize_cf_worlds = params["ML"][
      "VisualizeCfWorlds",
      "Whether the counterfactual worlds are visualized.", False]
    self._visualize_heatmap = params["ML"][
      "VisualizeCfHeatmap",
      "Whether the heatmap is being visualized.", False]
    self._results_folder = params["ML"][
      "ResultsFolder",
      "Whether the heatmap is being visualized.", "./"]
    self._logger = logging.getLogger()
    self._behavior_model_pool = behavior_model_pool or []
    self._ego_rule_based = ego_rule_based or BehaviorIDMLaneTracking(self._params)
    self._tracer = Tracer()
    if self._visualize_heatmap:
      _, self._axs_heatmap = plt.subplots(1, 1, constrained_layout=True)
    self._count = 0
    self._cf_axs = {}
    
  def reset(self, scenario=None):
    """resets the runtime and its objects"""
    self._count = 0
    return SingleAgentRuntime.reset(self, scenario=scenario)

  def ReplaceBehaviorModel(self, agent_id=None, behavior=None):
    """clones the world and replaced the behavior of an agent"""
    cloned_world = self._world.Copy()
    evaluators = self._evaluator._add_evaluators()
    for eval_key, eval in evaluators.items():
      cloned_world.AddEvaluator(eval_key, eval)
    if behavior is not None:
      cloned_world.agents[agent_id].behavior_model = behavior
    return cloned_world
  
  def GetAgentIds(self):
    """returns a list of the other agent's ids"""
    # NOTE: only use nearby agents
    agent_ids = list(self._world.agents.keys())
    # eval_id = self._scenario._eval_agent_ids[0]
    # agent_ids.remove(eval_id)
    return agent_ids

  def GenerateCounterfactualWorlds(self):
    """generates (len(agents) - 1) x M-behavior counterfactual worlds"""
    cf_worlds = []
    agent_ids = self.GetAgentIds()
    for agent_id in agent_ids:
      for behavior in self._behavior_model_pool:
        return_dict = {}
        return_dict[agent_id] = self.ReplaceBehaviorModel(agent_id, behavior)
        cf_worlds.append(return_dict)
    return cf_worlds

  def SimulateWorld(self, world, local_tracer, N=5, **kwargs):
    """simulates the world for N steps"""
    self.ml_behavior.set_action_externally = False
    eval_id = self._scenario._eval_agent_ids[0]
    self._world.agents[eval_id].behavior_model = self.ml_behavior
    replaced_agent_id = kwargs.get("replaced_agent", 0)
    if replaced_agent_id not in self._cf_axs and self._visualize_cf_worlds:
      self._cf_axs[replaced_agent_id] = {"ax": plt.subplots(3, 1, constrained_layout=True)[1], "count": 0}
    for i in range(0, N):
      if i == N - 1 and kwargs.get("num_virtual_world", 0) is not None and \
        self._visualize_cf_worlds and replaced_agent_id is not None:
        # NOTE: outsource
        for ftype in [".png", ".pgf"]:
          viewer = MPViewer(
            params=self._params,
            x_range=[-35, 35],
            y_range=[-35, 35],
            follow_agent_id=True,
            axis=self._cf_axs[replaced_agent_id]["ax"][self._cf_axs[replaced_agent_id]["count"]])
          # se
          for agent_id in world.agents.keys():
            viewer.agent_color_map[agent_id] = "gray"
          viewer.agent_color_map[replaced_agent_id] = (127/255, 205/255, 187/255)
          viewer.agent_color_map[eval_id] = (34/255, 94/255, 168/255)
          if replaced_agent_id == 1:
            viewer.drawWorld(
              world,
              eval_agent_ids=self._scenario._eval_agent_ids,
              filename=self._results_folder + "cf_%03d_replaced_" % self._count + str(replaced_agent_id)+ftype,
              debug_text=False)
        self._cf_axs[replaced_agent_id]["count"] += 1
      observed_world = world.Observe([eval_id])[0]
      eval_state = observed_world.Evaluate()
      agent_states = CaptureAgentStates(observed_world)
      eval_state = {**eval_state, **agent_states}
      # TODO: break at collision
      local_tracer.Trace(eval_state, **kwargs)
      if eval_state["collision"] or eval_state["drivable_area"]:
        break
      world.Step(self._step_time)
    self.ml_behavior.set_action_externally = True
  
  def St(self):
    self._start_time = time.time()

  def Et(self):
    end_time = time.time()
    dt = end_time - self._start_time
    self._logger.info(f"It took {dt:.3f} seconds to simulate all" + \
                      f" counterfactual worlds.")
  @property
  def tracer(self):
    return self._tracer
  
  def TraceCounterfactualWorldStats(self, local_tracer):
    collision_rate = local_tracer.collision_rate
    goal_reached = local_tracer.success_rate
    return {"collision": collision_rate,
            "goal_reached": goal_reached,
            "max_col_rate": self._max_col_rate}

  def FilterStates(self, states, **kwargs):
    states_ = []
    for state in states:
      for kwarg_key, kwarg_val in kwargs.items():
        if kwarg_key in state:
          if state[kwarg_key] == kwarg_val:
            states_.append(state)
    return states_

  def ExtractStatesPerWorld(self, states):
    pure_states_ = {}
    for state in states:
      world_idx = state["num_virtual_world"]
      pure_states_[world_idx] = []
      for key, item in state.items():
        if "state_" in key:
          pure_states_[world_idx].append(item)
      pure_states_[world_idx] = np.array(pure_states_[world_idx])
    return pure_states_

  def GetMeanForAgent(self, local_tracer, agent_id):
    filtered_states = self.FilterStates(local_tracer._states, replaced_agent=agent_id)
    extracted_states = self.ExtractStatesPerWorld(filtered_states)
    # print(extracted_states, agent_id)
    mean = None
    for k, v in extracted_states.items():
      if mean is None:
        mean = v
      else:
        mean += v
    mean /= len(extracted_states)
    return mean
      
  def DrawHeatmap(self, local_tracer, filename="./"):
    eval_id = self._scenario._eval_agent_ids[0]
    base_states = self.FilterStates(local_tracer._states, replaced_agent="None")
    extracted_base_states = self.ExtractStatesPerWorld(base_states)
    extracted_base_states_np = extracted_base_states[
      list(extracted_base_states.keys())[0]]
    
    # loop through all agents
    all_keys = list(local_tracer._states[0].keys())
    all_agent_ids = []
    for i, key in enumerate(all_keys):
      if "state_" in key:
        all_agent_ids.append(int(key.replace("state_", "")))
    # TODO: the ego agent is not replaced, but want influence
    arr = np.zeros(shape=(len(all_agent_ids), len(all_agent_ids)))
    for i, agent_id in enumerate(all_agent_ids):
      print(i, agent_id)
      mean = self.GetMeanForAgent(local_tracer, agent_id)
      row_from = np.sum((extracted_base_states_np - mean)**2, axis=1)
      arr[i, :] = row_from
    
    np.fill_diagonal(arr, 0.)
    self._axs_heatmap.imshow(arr, cmap=plt.get_cmap('Blues'))
    self._axs_heatmap.set_yticks(np.arange(len(all_agent_ids)))
    self._axs_heatmap.set_xticks(np.arange(len(all_agent_ids)))
    self._axs_heatmap.set_yticklabels(["$W^{v_"+str(agent_id)+"}$" for agent_id in all_agent_ids])
    self._axs_heatmap.set_xticklabels(["$\Delta_{v_"+str(agent_id)+"}$" for agent_id in all_agent_ids])
    self._axs_heatmap.set_rasterized(True)
    
    self._axs_heatmap.get_figure().savefig(filename+".png")
    self._axs_heatmap.get_figure().savefig(filename+".pgf")
    
  def step(self, action):
    """perform the cf evaluation"""
    # simulate counterfactual worlds
    local_tracer = Tracer()
    eval_id = self._scenario._eval_agent_ids[0]
    self.St()
    cf_worlds = self.GenerateCounterfactualWorlds()
    for v in self._cf_axs.values():
      v["count"] = 0
    for i, cf_world in enumerate(cf_worlds):
      cf_key = list(cf_world.keys())[0]
      self.SimulateWorld(
        cf_world[cf_key], local_tracer, N=self._cf_simulation_steps,
        replaced_agent=cf_key, num_virtual_world=i)
    self.Et()
    
    # NOTE: this world would actually have the predicted traj.
    gt_world = self.ReplaceBehaviorModel()
    self.SimulateWorld(
      gt_world, local_tracer, N=self._cf_simulation_steps,
      replaced_agent="None", num_virtual_world="None")
    # NOTE: outsource
    hist = gt_world.agents[eval_id].history
    traj = np.stack([x[0] for x in hist])
    self._viewer.drawTrajectory(traj, color='blue')
    
    if self._visualize_heatmap:
      self.DrawHeatmap(
        local_tracer,
        filename=self._results_folder + "cf_%03d" % self._count + "_heatmap")
  
    # evaluate counterfactual worlds
    trace = self.TraceCounterfactualWorldStats(local_tracer)
    collision_rate = trace['collision']/len(self._behavior_model_pool)
    print(collision_rate)
    self._logger.info(
      f"The counterfactual worlds have a collision" + \
      f"-rate of {collision_rate:.3f}.")

    # choose a policy
    executed_learned_policy = 1
    if collision_rate > self._max_col_rate:
      executed_learned_policy = 0
      self._logger.info(
        f"Executing fallback model.")
      self._world.agents[eval_id].behavior_model = self._ego_rule_based
    trace["executed_learned_policy"] = executed_learned_policy  
    self._tracer.Trace(trace)
    self._count += 1
    for fig in self._cf_axs.values():
      for sub_ax in fig["ax"]: 
        sub_ax.clear()
    return SingleAgentRuntime.step(self, action)


#!/usr/bin/env python


DELTA_TIMESTAMP_MS = 100  # similar throughout the whole dataset


class MotionState:
    def __init__(self, time_stamp_ms):
        assert isinstance(time_stamp_ms, int)
        self.time_stamp_ms = time_stamp_ms
        self.x = None
        self.y = None
        self.vx = None
        self.vy = None
        self.psi_rad = None

    def __str__(self):
        return "MotionState: " + str(self.__dict__)


class Track:
    def __init__(self, id):
        assert isinstance(id, int)
        self.track_id = id
        self.agent_type = None
        self.length = None
        self.width = None
        self.time_stamp_ms_first = None
        self.time_stamp_ms_last = None
        self.motion_states = dict()

    def __str__(self):
        string = "Track: track_id=" + str(self.track_id) + ", agent_type=" + str(self.agent_type) + \
               ", length=" + str(self.length) + ", width=" + str(self.width) + \
               ", time_stamp_ms_first=" + str(self.time_stamp_ms_first) + \
               ", time_stamp_ms_last=" + str(self.time_stamp_ms_last) + \
               "\n motion_states:"
        for key, value in sorted(self.motion_states.items()):
            string += "\n    " + str(key) + ": " + str(value)
        return string



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT

import numpy as np

from bark.core.models.behavior import BehaviorModel, \
    BehaviorMPContinuousActions, BehaviorMPMacroActions, BehaviorMacroActionsFromParamServer
from bark.core.models.dynamic import SingleTrackModel

from bark_ml.commons.py_spaces import Discrete


class BehaviorDiscreteMotionPrimitivesML(BehaviorMPContinuousActions):
  def __init__(self,
               params=None):
    BehaviorMPContinuousActions.__init__(
      self,
      params)
    self._min_max_acc = params["ML"]["BehaviorDiscretePrimitivesML"][
      "MinMaxAcc", "", [-3., 3.]]
    self._acc_d_steps = params["ML"]["BehaviorDiscretePrimitivesML"][
      "AccDiscretizationSteps", "", 10]
    self._min_max_steer = params["ML"]["BehaviorDiscretePrimitivesML"][
      "MinMaxSteeringRate", "", [-.2, .2]]
    self._steer_d_steps = params["ML"]["BehaviorDiscretePrimitivesML"][
      "SteeringRateDiscretizationSteps", "", 5]

    # add motion primitives
    for acc in np.linspace(
      self._min_max_acc[0], self._min_max_acc[1], self._acc_d_steps):
      for steering_rate in np.linspace(
        self._min_max_steer[0], self._min_max_steer[1], self._steer_d_steps):
        super().AddMotionPrimitive(
          np.array([acc, steering_rate], dtype=np.float32))

  @property
  def action_space(self):
    return Discrete(self.GetNumMotionPrimitives(None))


class BehaviorDiscreteMacroActionsML(BehaviorMPMacroActions):
  def __init__(self,
               params=None):
    self._macro_action_params = params.AddChild("ML").AddChild("BehaviorMPMacroActions")
    self._macro_action_params["BehaviorMPMacroActions"]["CheckValidityInPlan", "", False]
    default_model = BehaviorMacroActionsFromParamServer(self._macro_action_params)
    super().__init__(self._macro_action_params, default_model.GetMotionPrimitives())

  @property
  def action_space(self):
    return Discrete(len(self.GetMotionPrimitives()))


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

from abc import abstractmethod

class ConfigReader:
  def __init__(self, random_state, current_scenario_idx):
  # holds a global random state which should be used for sampling in config readers
    self.__random_state = random_state
  # holds current scenario idx if config reader adjust settings according to scenario idx ( e.g. databased models)
    self.__current_scenario_idx = current_scenario_idx

  @property
  def random_state(self):
    return self.__random_state

  @property
  def current_scenario_idx(self):
    return self.__current_scenario_idx


# The property-based scenario generation must implement all of this interfaces. Property readers only the respective interface
class ConfigReaderBehaviorModels(ConfigReader):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
  # returns list of size num agents with behavior models based on property, default_params_dict
  @abstractmethod
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    pass 
  
  # reimplement as param servers behavior models are serialized
  def get_param_servers(self):
    return []

class ConfigReaderExecutionModels(ConfigReader):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
  # returns list of size num agents with execution models based on property, default_params_dict
  @abstractmethod
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    pass

  # only reimplement if param server for this object shall be serialized (currently not needed)
  def get_param_servers(self):
    return []

class ConfigReaderDynamicModels(ConfigReader):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
  # returns list of size num agents with dynamic models based on property, default_params_dict
  @abstractmethod
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    pass

  # only reimplement if param server for this object shall be serialized (currently not needed)
  def get_param_servers(self):
    return []

class ConfigReaderAgentStatesAndGeometries(ConfigReader):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
  @abstractmethod
  # returns a tuple containing 1] list of size num agents each element being a python list representing an agent state
  # and 2] a list of size num agents of point lists defining the agent geometries
  #  3] None or a dict passed as kwargs to the other functions (
  # e.g. the dict can contain a type entry specifying for each agent if it is a truck or car)
  # and 4] a default_params_dict
  def create_from_config(self, config_param_object, road_corridor):
    pass

  # only reimplement if param server for this object shall be serialized (currently not needed)
  def get_param_servers(self):
    return []

class ConfigReaderDynamicModels(ConfigReader):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
  # returns list of size num agents with behavior models based on property, default_params_dict
  @abstractmethod
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    pass

  # only reimplement if param server for this object shall be serialized (currently not needed)
  def get_param_servers(self):
    return []


class ConfigReaderGoalDefinitions(ConfigReader):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
  # returns list of size num agents with goal definitions for each agent based on property, default_params_dict
  @abstractmethod
  def create_from_config(self, config_param_object, road_corridor, agent_states, controlled_agent_ids,  **kwargs):
    pass

  # only reimplement if param server for this object shall be serialized (currently not needed)
  def get_param_servers(self):
    return []

class ConfigReaderControlledAgents(ConfigReader):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
   # returns list of size num agents with true or false depending if agent is controlled or not for each agent based on property, default_params_dict
  @abstractmethod
  def create_from_config(self, config_param_object, road_corridor, agent_states,  **kwargs):
    pass

  # only reimplement if param server for this object shall be serialized (currently not needed)
  def get_param_servers(self):
    return []



from bark_ml.behaviors.cont_behavior import BehaviorContinuousML
from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMotionPrimitivesML
from bark_ml.behaviors.discrete_behavior import BehaviorDiscreteMacroActionsML


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Julian Bernhard, Klemens Esterle, Patrick Hart and
# Tobias Kessler
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import numpy as np
import glob
from collections import defaultdict

from bark.runtime.scenario.scenario_generation.config_readers.config_readers_interfaces \
       import ConfigReaderAgentStatesAndGeometries

from bark.runtime.scenario.interaction_dataset_processing import ShapeFromTrack, \
    BarkStateFromMotionState, InitStateFromTrack
from com_github_interaction_dataset_interaction_dataset.python.utils import dataset_reader

from bark.core.geometry.standard_shapes import *
from bark.core.geometry import *
from bark.core.models.dynamic import *

# this config reader defines agent states with distances sampled uniformly standard vehicle geometries models
# it can be specified with parameter "lane_position" being between 1 and num_road_corridor_lanes 
# in what lanes vehicles are placed, or if None in all lanes they are placed
class UniformVehicleDistribution(ConfigReaderAgentStatesAndGeometries):
  def create_from_config(self, config_param_object, road_corridor):
    self._lane_positions = config_param_object["LanePositions", "List of values out of 0:max_num_lanes-1. \
            Vehicles are placed only this lanes \
           or None if vehicles shall be placed at alles lanes", None]
    self._vehicle_distance_range = config_param_object["VehicleDistanceRange",
      "Distance range between vehicles in meter given as tuple from which" + \
      "distances are sampled uniformly",
      (10, 20)]
    self._other_velocity_range = config_param_object["OtherVehicleVelocityRange",
      "Lower and upper bound of velocity in km/h given as tuple from which" + \
      " velocities are sampled uniformly",
      (20,30)]
    self._s_range = config_param_object["SRange", "pair of values (tuple) between 0 and 1 to limit placement  \
                     to certain part of road corridor", (0.1, 0.7)]
    # todo make parameterizable, but not only based on 2d points
    self._vehicle_2d_shape = CarLimousine()

    agent_states = []
    agent_geometries = []
    agent_lane_positions = []

    lane_corridors, lane_positions = self.select_lane_corridors(road_corridor, self._lane_positions)
    for idx, lane_corridor in enumerate(lane_corridors):
      tmp_agent_states, tmp_agent_geometries = self.agents_along_lane_corridor(lane_corridor,
                   self._s_range[0], self._s_range[1])

      agent_states.extend(tmp_agent_states)
      agent_geometries.extend(tmp_agent_geometries)
      agent_lane_positions.extend([lane_positions[idx]]*len(tmp_agent_states))

    assert(len(agent_states) == len(agent_geometries))
    assert(len(agent_states) == len(agent_lane_positions))
    return agent_states, agent_geometries, {"agent_lane_positions": agent_lane_positions}, config_param_object


  def select_lane_corridors(self, road_corridor, lane_positions):
    lane_corridors = []
    if lane_positions:
      for lane_position in lane_positions:
        lane_corridors.append(road_corridor.lane_corridors[lane_position])
    else:
        lane_corridors = road_corridor.lane_corridors
        lane_positions = list(range(1, len(lane_corridors)+1))
    return lane_corridors, lane_positions

  def sample_velocity_uniform(self, velocity_range):
    return self.random_state.uniform(velocity_range[0], velocity_range[1])

  def sample_distance_uniform(self, distance_range):
    return self.random_state.uniform(distance_range[0], distance_range[1])


  def agents_along_lane_corridor(self,
                              lane_corridor,
                              s_start,
                              s_end):
    linestring = lane_corridor.center_line
    agent_states = []
    agent_geometries = []
    corridor_length = lane_corridor.center_line.Length()
    s = s_start*corridor_length
    while s < s_end*corridor_length:
      # set agent state on linestring with random velocity
      xy_point =  GetPointAtS(linestring, s)
      angle = GetTangentAngleAtS(linestring, s)
      
      velocity = self.sample_velocity_uniform(self._other_velocity_range)
      agent_state = [0, xy_point.x(), xy_point.y(), angle, velocity ]

      agent_states.append(agent_state)
      agent_geometries.append(self._vehicle_2d_shape)

      # move forward on linestring based on vehicle size and max/min distance
      s += self._vehicle_2d_shape.front_dist + self._vehicle_2d_shape.rear_dist + \
                  self.sample_distance_uniform(self._vehicle_distance_range)
    return agent_states, agent_geometries

class InteractionDataTrackIdsStatesGeometries(ConfigReaderAgentStatesAndGeometries):
  def create_from_config(self, config_param_object, road_corridor):
    track_file_name = config_param_object["TrackFilename", "Path to track file (csv)",
                                        "bark/runtime/tests/data/interaction_dataset_DEU_Merging_dummy_track.csv"]
    track_ids = config_param_object["TrackIds", "IDs of the vehicle tracks to import.", [1]]
    start_time = config_param_object["StartTs", "Timestamp when to start the scenario (ms)", 0]
    end_time = config_param_object["EndTs","Timestamp when to end the scenario (ms)", None]
    xy_offset = config_param_object["XYOffset", "offset in x and y direction.", [0, 0]]

    agent_geometries = []
    agent_states = []
    lane_positions = []
    tracks = []
    for track_id in track_ids:
      tracks_read = dataset_reader.read_tracks(track_file_name)
      track = tracks_read[track_id]
      if start_time is None:
          start_time = track.time_stamp_ms_first
      if end_time is None:
          end_time = track.time_stamp_ms_last
      numpy_state = InitStateFromTrack(track, xy_offset, start_time)
      agent_state = numpy_state.reshape(5).tolist()
      agent_states.append(agent_state)

      shape = ShapeFromTrack(track)
      agent_geometries.append(shape)

      tracks.append(track)

      lane_positions_agent = self.find_lane_positions(numpy_state, road_corridor)
      lane_positions.append(lane_positions_agent)

    assert(len(agent_states) == len(agent_geometries))
    return agent_states, agent_geometries, {"track_ids": track_ids, "tracks" : tracks, \
             "agent_ids" : track_ids, "xy_offset" : xy_offset, "start_time" : start_time, "end_time" : end_time, \
               "agent_lane_positions" : lane_positions}, config_param_object
  
  def find_lane_positions(self, init_state, road_corridor):
    lps = []
    for idx, lane_corridor in enumerate(road_corridor.lane_corridors):
      if Collide(lane_corridor.polygon, Point2d(init_state[int(StateDefinition.X_POSITION)], \
         init_state[int(StateDefinition.Y_POSITION)])):
        lps.append(idx)
    return lps

class InteractionDataWindowStatesGeometries(ConfigReaderAgentStatesAndGeometries):
  window_start = None
  window_end = None
  track_dict_list = None
  current_track_file = None
  def create_from_config(self, config_param_object, road_corridor):
    track_file_names = config_param_object["TrackFilenames", "Path to track file (csv)",
                                        "bark/runtime/tests/data/*_dataset_dummy_track.csv"]
    xy_offset = config_param_object["XYOffset", "offset in x and y direction.", [0, 0]]
    wheel_base = config_param_object["WheelBase", "Wheelbase assumed for shape calculation", 2.7]
    window_length = config_param_object["WindowLength", "Window length for search of agents for a scenario ", 200]
    skip_time_scenarios = config_param_object["SkipTimeScenarios", "Time delta between start of previous scenario window and next init of search window", 0]
    skip_time_search = config_param_object["SkipTimeSearch", "Time delta between start of current and next search window", 100]
    min_time = config_param_object["MinTime", "Time offset from beginning of track file to start searching", 0]
    max_time = config_param_object["MaxTime", "Max time included in search", 100000]
    only_on_one_lane = config_param_object["OnlyOnOneLane", "If True only scenarios are defined where agents are on a single lane", False]
    minimum_numbers_per_lane = config_param_object["MinimumNumbersPerLane", "List where each element specifies how man vehicles must be at minimum at this lane,\
                                  lane position equals list index", [1, 0]]

    window_start = InteractionDataWindowStatesGeometries.window_start 
    window_end = InteractionDataWindowStatesGeometries.window_end
    track_dict_list = InteractionDataWindowStatesGeometries.track_dict_list
    current_track_file = InteractionDataWindowStatesGeometries.current_track_file
    # reset when a new scenario generation starts
    if self.current_scenario_idx == 0:
      window_start = min_time
      window_end = min_time + window_length
      track_dict_list = self.load_tracks(track_file_names)
      current_track_file = 0
    else:
      window_start += skip_time_scenarios
      window_end += skip_time_scenarios

    scenario_track_ids = []
    while len(scenario_track_ids) == 0 and current_track_file < len(track_dict_list):
      track_dict = track_dict_list[current_track_file]
      scenario_track_ids, window_start, window_end = self.find_track_ids_moving_window(xy_offset, window_start, window_end, track_dict, only_on_one_lane, minimum_numbers_per_lane,
                        window_length, skip_time_search, min_time, max_time, road_corridor, wheel_base)
      if len(scenario_track_ids) > 0:
        break
      window_start = min_time
      window_end = min_time + window_length
      current_track_file += 1
    if len(scenario_track_ids) < 1:
      raise ValueError("No track ids found for scenario idx {}. Consider lowering the number of scenarios.".format(self.current_scenario_idx))

    agent_geometries = []
    agent_states = []
    lane_positions = []
    tracks = []
    for track_id in scenario_track_ids:
      track = track_dict[track_id]
      numpy_state = self.get_init_state(track, xy_offset, window_start)
      agent_state = numpy_state.reshape(5).tolist()
      agent_states.append(agent_state)
      shape = ShapeFromTrack(track)
      agent_geometries.append(shape)
      tracks.append(track)
      lane_positions_single = self.find_lane_positions(road_corridor, self.get_shape_at_time_point(track, xy_offset, window_start))
      lane_positions.append(lane_positions_single)

    assert(len(agent_states) == len(agent_geometries))
    InteractionDataWindowStatesGeometries.window_start = window_start
    InteractionDataWindowStatesGeometries.window_end = window_end
    InteractionDataWindowStatesGeometries.track_dict_list = track_dict_list
    InteractionDataWindowStatesGeometries.current_track_file = current_track_file

    return agent_states, agent_geometries, {"track_ids": scenario_track_ids, "tracks" : tracks, \
             "agent_ids" : scenario_track_ids, "xy_offset" : xy_offset, "start_time" : window_start, "end_time" : window_end, \
               "agent_lane_positions" : lane_positions}, config_param_object

  def load_tracks(self, track_filenames):
    track_dict_list = []
    if isinstance(track_filenames, list):
      track_files_found = track_filenames
    else:
      track_files_found = glob.glob(track_filenames)
    for filename in track_files_found:
      track_dict = dataset_reader.read_tracks(filename)
      track_dict_list.append(track_dict)
    return track_dict_list

  def find_track_ids_moving_window(self, xy_offset, window_start, window_end, track_dict, only_on_one_lane, minimum_numbers_per_lane, \
                                      window_length, skip_time_search, time_offset, max_time, road_corridor, wheel_base):
    def move_window(window_start, window_end):
      window_start += skip_time_search
      window_end = window_start + window_length
      return window_start, window_end

    while True:
      window_start, window_end = move_window(window_start, window_end)
      if window_end > max_time:
        return [], window_start, window_end
      window_track_ids = self.find_track_ids(track_dict, window_start, window_end)
      if len(window_track_ids) < 1:
        continue

      numbers_per_lane = defaultdict(list)
      for track_id in window_track_ids:
        track = track_dict[track_id]
        shape = self.get_shape_at_time_point(track, xy_offset, window_start)
        lane_positions = self.find_lane_positions(road_corridor, shape)
        # skip whole window if lane positions not fulfilled
        if only_on_one_lane and len(lane_positions) != 1:
          continue
        if len(lane_positions) == 0:
          continue
        if not Within(shape, road_corridor.polygon):
          continue
        numbers_per_lane[lane_positions[0]].append(track_id)

      valid_track_ids = []
      desired_number_valid = True
      for lane_pos, minimum_number in enumerate(minimum_numbers_per_lane):
        if len(numbers_per_lane[lane_pos]) < minimum_number:
          desired_number_valid = False
          break
        else:
          valid_track_ids.extend(numbers_per_lane[lane_pos])
      if desired_number_valid:
        break
      else:
        continue

    return valid_track_ids, window_start, window_end

  def get_init_state(self, track, xy_offset, start_time):
    return InitStateFromTrack(track, xy_offset, start_time)

  def find_track_ids(self, track_dict, start_time, end_time):
    list_ids = []
    for id_current in track_dict.keys():
        if track_dict[id_current].time_stamp_ms_last >= end_time and \
          track_dict[id_current].time_stamp_ms_first <= start_time:
            list_ids.append(id_current)
    return list_ids

  def get_shape_at_time_point(self, track, xy_offset, time_point):
    init_state = self.get_init_state(track, xy_offset, time_point)
    shape = ShapeFromTrack(track)
    shape_at_timepoint = shape.Transform([init_state[int(StateDefinition.X_POSITION)], \
                                   init_state[int(StateDefinition.Y_POSITION)], \
                                   init_state[int(StateDefinition.THETA_POSITION)]])
    return shape_at_timepoint

  def find_lane_positions(self, road_corridor, shape):
    lps = []
    for idx, lane_corridor in enumerate(road_corridor.lane_corridors):
      if Collide(lane_corridor.polygon, shape):
        lps.append(idx)
    return lps


# Copyright (c) 2020 fortiss GmbH
#
# Authors: Marco Oliva
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

import gin
import tensorflow as tf # pylint: disable=g-explicit-tensorflow-version-import

from tf_agents.agents.sac import sac_agent
from tf_agents.networks import network, normal_projection_network, encoding_network
from tf_agents.utils import nest_utils
from bark.runtime.commons.parameters import ParameterServer
from bark_ml.library_wrappers.lib_tf_agents.networks.gnns.interaction_wrapper import InteractionWrapper


def projection_net(spec):
  return normal_projection_network.NormalProjectionNetwork(
    spec,
    mean_transform=None,
    state_dependent_std=True,
    init_means_output_factor=0.1,
    std_transform=sac_agent.std_clip_transform,
    scale_distribution=True)

@gin.configurable
class GNNActorNetwork(network.Network):
  """Creates an actor GNN."""

  def __init__(self,
               input_tensor_spec,
               output_tensor_spec,
               gnn,
               fc_layer_params=None,
               dropout_layer_params=None,
               conv_layer_params=None,
               activation_fn=tf.nn.relu,
               name='ActorNetwork',
               params=ParameterServer()):
    """Creates an instance of `ActorNetwork`.
    Args:
      input_tensor_spec: A nest of `tensor_spec.TensorSpec` representing the
        inputs.
      output_tensor_spec: A nest of `tensor_spec.BoundedTensorSpec` representing
        the outputs.
      gnn: The function that initializes a graph neural network that 
        accepts the input observations and computes node embeddings.
      fc_layer_params: Optional list of fully_connected parameters, where each
        item is the number of units in the layer.
      dropout_layer_params: Optional list of dropout layer parameters, each item
        is the fraction of input units to drop or a dictionary of parameters
        according to the keras.Dropout documentation. The additional parameter
        `permanent', if set to True, allows to apply dropout at inference for
        approximated Bayesian inference. The dropout layers are interleaved with
        the fully connected layers; there is a dropout layer after each fully
        connected layer, except if the entry in the list is None. This list must
        have the same length of fc_layer_params, or be None.
      conv_layer_params: Optional list of convolution layers parameters, where
        each item is a length-three tuple indicating (filters, kernel_size,
        stride).
      activation_fn: Activation function, e.g. tf.nn.relu, slim.leaky_relu, ...
      name: A string representing name of the network.
    Raises:
      ValueError: If `input_tensor_spec` or `action_spec` contains more than one
        item, or if the action data type is not `float`.
    """
    super(GNNActorNetwork, self).__init__(
      input_tensor_spec=input_tensor_spec,
      state_spec=(),
      name=name)

    if len(tf.nest.flatten(input_tensor_spec)) > 1:
      raise ValueError('Only a single observation is supported by this network')
    
    flat_action_spec = tf.nest.flatten(output_tensor_spec)
    
    if len(flat_action_spec) > 1:
      raise ValueError('Only a single action is supported by this network')
    
    if flat_action_spec[0].dtype not in [tf.float32, tf.float64]:
      raise ValueError('Only float actions are supported by this network.')
    
    if gnn is None:
      raise ValueError('`gnn` must not be `None`.')

    self._gnn = gnn(name=name, params=params)
    self._latent_trace = None
    self._encoder = encoding_network.EncodingNetwork(
      input_tensor_spec=tf.TensorSpec([None, self._gnn._embedding_size]),
      preprocessing_layers=None,
      preprocessing_combiner=None,
      conv_layer_params=conv_layer_params,
      fc_layer_params=fc_layer_params,
      dropout_layer_params=dropout_layer_params,
      activation_fn=activation_fn,
      kernel_initializer=tf.compat.v1.keras.initializers.glorot_uniform(),
      batch_squash=False,
      dtype=tf.float32)

    self._projection_nets = tf.nest.map_structure(projection_net, output_tensor_spec)
    self._output_tensor_spec = tf.nest.map_structure(lambda proj_net: proj_net.output_spec,
                                        self._projection_nets)

  @property  
  def output_tensor_spec(self):
    return self._output_tensor_spec

  def call(self, observations, step_type=(), network_state=(), training=False):
    del step_type # unused.

    if len(observations.shape) == 1:
      observations = tf.expand_dims(observations, axis=0)
    batch_size = tf.shape(observations)[0]
    embeddings = self._gnn(observations, training=training)
    self._latent_trace = self._gnn._latent_trace

    # extract ego state (node 0)
    if batch_size > 0: 
      embeddings = embeddings[:, 0]

    # with tf.name_scope("GNNActorNetwork"):
    #   tf.summary.histogram("actor_gnn_output", embeddings)
    
    output, network_state = self._encoder(embeddings, training=training)
    output = embeddings
    
    outer_rank = nest_utils.get_outer_rank(
      observations, 
      self.input_tensor_spec)

    def call_projection_net(net):
      distribution, _ = net(output, outer_rank, training=training)
      return distribution

    output_actions = tf.nest.map_structure(
      call_projection_net, self._projection_nets)

    return output_actions, network_state



# Copyright (c) 2020 fortiss GmbH
#
# Authors: Patrick Hart, Julian Bernhard, Klemens Esterle, and
# Tobias Kessler
#
# This software is released under the MIT License.
# https://opensource.org/licenses/MIT


class Blueprint:
  def __init__(self,
               scenario_generation=None,
               viewer=None,
               dt=None,
               evaluator=None,
               observer=None,
               ml_behavior=None):
    self._scenario_generation = scenario_generation
    self._viewer = viewer
    self._dt = dt
    self._evaluator = evaluator
    self._observer = observer
    self._ml_behavior = ml_behavior
